use std::{fs::{self, File}, path::PathBuf, time::SystemTime};

use anyhow::Context;
use egui::ahash::HashMap;
use serde::{Deserialize, Serialize};
use serde;

use crate::{mod_builder::ModConfiguration, omori_locator};

#[derive(Deserialize, Serialize, Debug)]
struct Config {
    key: Option<Vec<u8>>,

    #[serde(default)]
    recently_used_game_paths: HashMap<PathBuf, u64>,

    #[serde(default)]
    recently_used_playtest_paths: HashMap<PathBuf, u64>,

    #[serde(default)]
    configuration_history: HashMap<PathBuf, ModConfiguration>
}

impl Default for Config {
    fn default() -> Self {
        Self { 
            key: Default::default(), 
            recently_used_game_paths: Default::default(),
            recently_used_playtest_paths: Default::default(),
            configuration_history: Default::default()
        }
    }
}

#[derive(Debug)]
pub struct ConfigProvider {
    config: Config,
    base_path: PathBuf
}

impl ConfigProvider {
    pub fn new() -> anyhow::Result<ConfigProvider> {
        let mut base_path = dirs::preference_dir().with_context(|| "Failed to find your preference directory")?;
        base_path.push(env!("CARGO_PKG_NAME"));

        fs::create_dir_all(&base_path)?;

        base_path.push("config.json");

        if !fs::exists(&base_path)? {
            let mut fd = File::create(&base_path)?;
            serde_json::to_writer(&mut fd, &Config::default())?;
            drop(fd);
        }

        let mut fd = File::open(&base_path)?;
        let config: Config = serde_json::from_reader(&mut fd)?;
        drop(fd);

        Ok(ConfigProvider { config, base_path })
    }

    fn commit(&mut self) -> anyhow::Result<()> {
        let mut fd = File::create(&self.base_path)?;
        serde_json::to_writer(&mut fd, &self.config)?;
        drop(fd);

        Ok(())
    }

    pub fn get_key(&self) -> Option<Vec<u8>> {
        return self.config.key.clone();
    }

    pub fn set_key(&mut self, key: &Vec<u8>) -> anyhow::Result<()> {
        self.config.key = Some(key.clone());
        self.commit()?;

        Ok(())
    }

    pub fn get_game_paths(&self) -> anyhow::Result<Vec<(PathBuf, u64)>> {
        let mut paths: Vec<(&PathBuf, &u64)> = self.config.recently_used_game_paths.iter().collect();
        paths.sort_by(|a, b| a.1.cmp(b.1).reverse());
        
        Ok(paths
            .iter()
            .filter(|p| omori_locator::validate_omori_installation(p.0))
            .map(|v| (v.0.clone(), *v.1))
            .collect()
        )
    }

    pub fn set_game_path_used(&mut self, which: &PathBuf) -> anyhow::Result<()> {
        let current_time = SystemTime::now().duration_since(SystemTime::UNIX_EPOCH)?.as_secs();
        self.config.recently_used_game_paths.insert(which.clone(), current_time);
        self.config.recently_used_game_paths = self.config.recently_used_game_paths
            .clone()
            .into_iter()
            .filter(|p| omori_locator::validate_omori_installation(&p.0))
            .collect();
        self.commit()?;

        Ok(())
    }

    pub fn get_playtest_paths(&self) -> anyhow::Result<Vec<(PathBuf, u64)>> {
        let mut paths: Vec<(&PathBuf, &u64)> = self.config.recently_used_playtest_paths.iter().collect();
        paths.sort_by(|a, b| a.1.cmp(b.1).reverse());
        
        Ok(paths
            .iter()
            .filter(|p| omori_locator::validate_playtest(p.0))
            .map(|v| (v.0.clone(), *v.1))
            .collect()
        )
    }

    pub fn set_playtest_path_used(&mut self, which: &PathBuf) -> anyhow::Result<()> {
        let current_time = SystemTime::now().duration_since(SystemTime::UNIX_EPOCH)?.as_secs();
        self.config.recently_used_playtest_paths.insert(which.clone(), current_time);
        self.config.recently_used_playtest_paths = self.config.recently_used_playtest_paths
            .clone()
            .into_iter()
            .filter(|p| omori_locator::validate_playtest(&p.0))
            .collect();
        self.commit()?;

        Ok(())
    }

    pub fn get_configuration_for_playtest(&self, which: &PathBuf) -> ModConfiguration {
        match self.config.configuration_history.get(which) {
            Some(v) => v.clone(),
            None => Default::default()
        }
    }

    pub fn set_configuration_for_playtest(&mut self, which: &PathBuf, config: &ModConfiguration) -> anyhow::Result<()> {
        self.config.configuration_history.insert(which.clone(), config.clone());
        self.commit()?;

        Ok(())
    }
}
