use anyhow::Context;
use log::{debug, info, trace};
use regex::bytes::Regex;
use sha2::Digest;
use std::{fs::{self, File}, io::Read, path::PathBuf};

#[cfg(any(target_os = "linux", target_os = "macos"))]
fn get_steam_root() -> anyhow::Result<PathBuf> {
    let mut base = dirs::home_dir().with_context(|| "Couldn't figure out your home directory")?;

    #[cfg(target_os="linux")]
    base.push(".steam/root");

    #[cfg(target_os="macos")]
    base.push("Library/Application Support/Steam");

    info!("Using {:?} as steam base", base);

    Ok(base)
}

#[cfg(target_os = "windows")]
fn get_steam_root() -> anyhow::Result<PathBuf> {
    use registry::{self, Security, Data};
    
    let key = registry::Hive::LocalMachine.open(r"SOFTWARE\Wow6432Node\Valve\Steam", Security::Read)?;
    let value = match key.value("InstallPath")? {
        Data::String(s) => Some(s.to_string_lossy()),
        Data::ExpandString(s) => Some(s.to_string_lossy()),
        _ => None
    }.with_context(|| "Couldn't read the steam install path from registry")?;

    let mut base = PathBuf::from(value);

    debug!("Using {:?} as steam base", base);

    Ok(base)
}

const OMORI_STEAM_APPID: &str = "1150690";

pub fn get_omori_path() -> anyhow::Result<PathBuf> {
    let mut base = get_steam_root()?;

    base.push("steamapps");
    base.push("libraryfolders.vdf");
    
    info!("Searching {:?} for library folders with omori", base);

    let mut buffer = Vec::new();

    let mut fd = File::open(base)?;
    fd.read_to_end(&mut buffer)?;

    let data = String::from_utf8(buffer)?;
    let data = keyvalues_parser::Vdf::parse(&data.as_str())?;
    let data = data.value.get_obj().with_context(|| "Couldn't get the steam library libraryfolders object.")?;
    let keys = data.keys();

    let mut omori_path: Option<PathBuf> = None;
    for key in keys {
        let entry = data.get(key).with_context(|| "Couldn't read specific steam library libraryfolders object.")?;
        let entry = entry.get(0).with_context(|| "Couldn't read specific steam library libraryfolders object.")?;
        let entry = entry.get_obj().with_context(|| "libraryfolders parsing failed: Not an object")?;

        let apps = entry.get("apps").with_context(|| "libraryfolders parsing failed: No apps")?;
        let apps = apps.get(0).with_context(|| "libraryfolders parsing failed: No apps")?;
        let apps = apps.get_obj().with_context(|| "libraryfolders parsing failed: No apps")?;
        let apps: Vec<_> = apps.keys().collect();

        let path = entry.get("path").with_context(|| "libraryfolders parsing failed: No path")?;
        let path = path.get(0).with_context(|| "libraryfolders parsing failed: No path")?;
        let path = path.get_str().with_context(|| "libraryfolders parsing failed: No path")?;

        if apps.iter().find(|x| (**x).eq(OMORI_STEAM_APPID)).is_some() {
            let mut base = PathBuf::from(path);

            #[cfg(not(target_os = "macos"))]
            {
                base.push("steamapps");
                base.push("common");
                base.push("OMORI");
                base.push("www");
            }

            #[cfg(target_os = "macos")]
            {
                base.push("steamapps");
                base.push("common");
                base.push("OMORI");
                base.push("OMORI.app");
                base.push("Contents");
                base.push("Resources");
                base.push("app.nw");
            }
            
            omori_path = Some(base);
        }
    }

    
    let path = omori_path.with_context(|| "Failed to find omori")?;
    debug!("{:?} seems to be it.", path);
    if validate_omori_installation(&path) {
        Ok(path)
    } else {
        Err(anyhow::anyhow!("The steam installation of OMORI is not valid"))
    }
}

const GAME_KEY_HASH: &[u8; 32] = include_bytes!("keyhash");

pub fn get_omori_key() -> anyhow::Result<Vec<u8>> {
    let mut base: PathBuf = get_steam_root()?;

    base.push("appcache");
    base.push("appinfo.vdf");

    let mut buffer = Vec::new();

    let mut fd = File::open(base)?;
    fd.read_to_end(&mut buffer)?;
    drop(fd);

    let re = Regex::new(r"--([0-9a-f]{32})")?;

    for (_, [key]) in re.captures_iter(&buffer).map(|c| c.extract()) {
        let hash = sha2::Sha256::digest(key);

        let mut equal = true;
        for i in 0..32 {
            if hash[i] != GAME_KEY_HASH[i] {
                equal = false;
            }
        }
        
        if equal {
            return Ok(key.to_vec())
        }
    }


    Err(anyhow::anyhow!("Couldn't find any valid decryption key for OMORI in your Steam installation."))
}


const EXPECTED_PATHS_BASE: [&str; 32] = [
    "index.html",
    "editor.json",
    "audio/bgm/AMB_forest.rpgmvo",
    "audio/me/cutscene_omori_saves_player.rpgmvo",
    "data/Actors.KEL",
    "data/System.KEL",
    "data/Atlas.PLUTO",
    "data/Troops.KEL",
    "fonts/OMORI_GAME.ttf",
    "icon/icon.png",
    "img/animations/e_calm_down.rpgmvp",
    "img/atlases/battleATLAS.rpgmvp",
    "img/battlebacks1/battleback_dw_humphrey.rpgmvp", // best area
    "img/characters/!FA_OBJECTS_2.rpgmvp",
    "img/characters/FA_AUBREY.rpgmvp",
    "img/enemies/!battle_snow_bunny.rpgmvp",
    "img/faces/01_OMORI_BATTLE.rpgmvp",
    "img/overlays/fog.rpgmvp",
    "img/parallaxes/!BS_faces.rpgmvp",
    "img/pictures/basil_something.rpgmvp",
    "img/slotmachine/bet_line_2.rpgmvp",
    "img/sv_actors/!battle_aubrey.rpgmvp",
    "img/system/bar_gradients.rpgmvp",
    "img/system/Window.png",
    "img/tilesets/BS_Beach.rpgmvp",
    "img/transition/starburst_2.rpgmvp",
    "js/rpg_core.js",
    "js/plugins/Omori BASE.OMORI",
    "languages/en/10_cutscenes_fakeknifefight.HERO",
    "maps/FA_SUNSET_INTERIOR2.AUBREY",
    "movies/secret_ending.webm",
    "movies/Bad End Credits Draft 4.webm"
];

pub fn validate_omori_installation(base: &PathBuf) -> bool {
    info!("Validating {:?}", base);
    for path in EXPECTED_PATHS_BASE {
        trace!("Checking {path}");
        let mut real = base.clone();
        real.push(path);
        
        if let Ok(r) = fs::exists(&real) {
            if !r { return false; }
        } else { return false; }
    }
    info!("Validation passed");

    true
}

const EXPECTED_PATHS_PLAYTEST: [&str; 17] = [
    "index.html",
    "Game.rpgproject",
    "img",
    "img/pictures",
    "audio",
    "audio/bgm",
    "js",
    "languages",
    "js/plugins",
    "js/plugins/Omori BASE.js",
    "package.json",
    "data/Actors.json",
    "data/System.json",
    "data/Atlas.yaml",
    "data/Troops.json",
    "img/system/Window.png",
    "img/atlases/battleATLAS.png",
];

pub fn validate_playtest(base: &PathBuf) -> bool {
    info!("Validating {:?}", base);
    for path in EXPECTED_PATHS_PLAYTEST {
        trace!("Checking {path}");
        let mut real = base.clone();
        real.push(path);
        
        if let Ok(r) = fs::exists(&real) {
            if !r { return false; }
        } else { return false; }
    }
    info!("Validation passed");

    true
}